#![no_std]
#![no_main]

mod button;
mod crypto;
mod datetime;
mod screen;
mod storage;

use arduino_hal::default_serial;
use arduino_hal::delay_ms;
// use arduino_hal::delay_ms;
use arduino_hal::prelude::*;
use crypto::generate_otp_token;
use crypto::INTERVAL;
use ds323x::Rtcc;
use nb::block;
use panic_halt as _;
use storage::SECRET_KEY_MAX_LEN;
use storage::SECRET_KEY_NAME_LEN;

const ENDL: u8 = 0;
const OK: u8 = 1;
const ERROR: u8 = 2;
const HANDSHAKE: u8 = 3;
const SET_TIMESTAMP: u8 = 10;
const ADD_TOKEN: u8 = 20;
const DELETE_TOKEN: u8 = 30;
const GET_TOKENS: u8 = 40;
const WIPE_TOKENS: u8 = 50;
const EXIT: u8 = 254;

#[arduino_hal::entry]
fn main() -> ! {
    let dp = arduino_hal::Peripherals::take().unwrap();
    let pins = arduino_hal::pins!(dp);

    let mut led = pins.d13.into_output();
    let mut serial = default_serial!(dp, pins, 9600);

    let i2c_clock = arduino_hal::I2c::new(
        dp.TWI,
        pins.a4.into_pull_up_input(),
        pins.a5.into_pull_up_input(),
        50000,
    );
    let mut rtc = ds323x::Ds323x::new_ds3231(i2c_clock);

    let mut rs = pins.d12.into_output();
    let mut en = pins.d11.into_output();
    let mut d4 = pins.d5.into_output();
    let mut d5 = pins.d4.into_output();
    let mut d6 = pins.d3.into_output();
    let mut d7 = pins.d2.into_output();

    let mut display = screen::StnScreen::new(&mut rs, &mut en, &mut d4, &mut d5, &mut d6, &mut d7);
    display.clear();

    let datetime = datetime::Datetime::from_ds3231(&mut rtc);

    let mut eeprom = arduino_hal::Eeprom::new(dp.EEPROM);
    let mut tokens = storage::Tokens::new(&mut eeprom, datetime.unix_epoch());

    let up = pins.d6.into_pull_up_input();
    let mut button = button::Button::new(&up, true);

    button.update();
    if button.update() == button::Event::Pressed {
        display.write_str("Connected to");
        display.set_cursor(0, 1);
        display.write_str("USB...");
        loop {
            // Waiting from a command since the tool
            let cmd = block!(serial.read()).unwrap_or(u8::MAX);
            display.set_cursor(8, 1);
            display.write_str("---");
            display.set_cursor(8, 1);
            display.write_u32(cmd as u32);

            match cmd {
                EXIT => {}
                HANDSHAKE => serial.write(OK).unwrap(),
                SET_TIMESTAMP => {
                    serial.write(OK).unwrap();
                    let year = loop {
                        let y = block!(serial.read()).unwrap();
                        serial.write(y).unwrap();
                        if block!(serial.read()).unwrap() == OK {
                            break y;
                        }
                    };
                    let month = loop {
                        let m = block!(serial.read()).unwrap();
                        serial.write(m).unwrap();
                        if block!(serial.read()).unwrap() == OK {
                            break m;
                        }
                    };
                    let day = loop {
                        let d = block!(serial.read()).unwrap();
                        serial.write(d).unwrap();
                        if block!(serial.read()).unwrap() == OK {
                            break d;
                        }
                    };
                    let hours = loop {
                        let h = block!(serial.read()).unwrap();
                        serial.write(h).unwrap();
                        if block!(serial.read()).unwrap() == OK {
                            break h;
                        }
                    };
                    let minutes = loop {
                        let m = block!(serial.read()).unwrap();
                        serial.write(m).unwrap();
                        if block!(serial.read()).unwrap() == OK {
                            break m;
                        }
                    };
                    let seconds = loop {
                        let s = block!(serial.read()).unwrap();
                        serial.write(s).unwrap();
                        if block!(serial.read()).unwrap() == OK {
                            break s;
                        }
                    };
                    rtc.set_year(year as u16 + 2000).unwrap();
                    rtc.set_month(month).unwrap();
                    rtc.set_day(day).unwrap();
                    rtc.set_hours(ds323x::Hours::H24(hours)).unwrap();
                    rtc.set_minutes(minutes).unwrap();
                    rtc.set_seconds(seconds).unwrap();
                    serial.write(OK).unwrap();
                }
                ADD_TOKEN => match tokens.search_free() {
                    Some(index) => {
                        serial.write(OK).unwrap();
                        let mut name_buff = [0u8; SECRET_KEY_NAME_LEN as usize];
                        for i in 0..name_buff.len() {
                            let ch = block!(serial.read()).unwrap();
                            if ch == ENDL {
                                break;
                            } else {
                                name_buff[i] = ch;
                            }
                        }

                        let mut key_buff = [0u8; SECRET_KEY_MAX_LEN as usize];
                        for i in 0..key_buff.len() {
                            let ch = block!(serial.read()).unwrap();
                            if ch == ENDL {
                                break;
                            } else {
                                key_buff[i] = ch;
                            }
                        }
                        match tokens.write(index, &name_buff, &key_buff) {
                            Some(_) => serial.write(OK).unwrap(),
                            None => serial.write(ERROR).unwrap(),
                        }
                    }
                    None => serial.write(ERROR).unwrap(),
                },
                WIPE_TOKENS => {
                    serial.write(OK).unwrap();
                    tokens.wipe_all_tokens();
                    serial.write(OK).unwrap();
                }
                _ => {}
            }
        }
    }

    match tokens.current {
        Some(index) => {
            let mut buff_name = [0u8; SECRET_KEY_NAME_LEN as usize];
            let mut buff_key = [0u8; SECRET_KEY_MAX_LEN as usize];
            match tokens.read(index, &mut buff_name, &mut buff_key) {
                Some((len_name, len_key)) => {
                    let name = core::str::from_utf8(&buff_name[..len_name]).unwrap();
                    let key = &buff_key[..len_key];
                    let dt = datetime::Datetime::from_ds3231(&mut rtc);

                    let timestamp = dt.unix_epoch();

                    let token = generate_otp_token(key, timestamp);

                    display.set_cursor(0, 0);
                    display.write_str(name);
                    display.set_cursor(0, 1);
                    display.write_u32(token);
                }
                _ => {
                    // display.clear();
                    // display.set_cursor(0, 0);
                    // display.write_str("ERROR");
                }
            }
        }
        _ => {}
    }

    loop {
        led.toggle();
        delay_ms(1000);
    }
}
